"""Voice over IP (VoIP) implementation."""
import asyncio
import logging
import socket
import struct
import time
from abc import ABC, abstractmethod
from dataclasses import dataclass
from functools import partial
from typing import Any, Callable, Optional, Set

from .const import OPUS_PAYLOAD_TYPE
from .rtp_audio import RtpOpusInput, RtpOpusOutput
from .sip import CallInfo, SdpInfo, SipDatagramProtocol

_LOGGER = logging.getLogger(__name__)
_RTCP_BYE = 203


@dataclass
class RtcpState:
    """State of a call according to RTCP packets received."""

    bye_callback: Optional[Callable[[], None]] = None


CallProtocolFactory = Callable[[CallInfo, RtcpState], asyncio.Protocol]


class VoipDatagramProtocol(SipDatagramProtocol):
    """UDP server for Voice over IP (VoIP)."""

    def __init__(
        self,
        sdp_info: SdpInfo,
        valid_protocol_factory: CallProtocolFactory,
        invalid_protocol_factory: Optional[CallProtocolFactory] = None,
    ) -> None:
        """Set up VoIP call handler."""
        super().__init__(sdp_info)
        self.valid_protocol_factory = valid_protocol_factory
        self.invalid_protocol_factory = invalid_protocol_factory
        self._tasks: Set[asyncio.Future[Any]] = set()

    def is_valid_call(self, call_info: CallInfo) -> bool:
        """Filter calls."""
        return True

    def on_call(self, call_info: CallInfo):
        """Answer incoming calls and start RTP server on a random port."""
        protocol_factory = (
            self.valid_protocol_factory
            if self.is_valid_call(call_info)
            else self.invalid_protocol_factory
        )
        if protocol_factory is None:
            _LOGGER.debug("Call rejected: %s", call_info)
            return

        rtp_ip = ""
        if call_info.local_rtp_port is None:
            # Find free RTP/RTCP ports
            rtp_port = 0

            while True:
                sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
                sock.setblocking(False)

                # Bind to a random UDP port
                sock.bind(("", 0))
                rtp_ip, rtp_port = sock.getsockname()

                # Close socket to free port for re-use
                sock.close()

                # Check that the next port up is available for RTCP
                sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
                try:
                    sock.bind(("", rtp_port + 1))

                    # Will be opened again below
                    sock.close()

                    # Found our ports
                    break
                except OSError:
                    # RTCP port is taken
                    pass

        else:
            rtp_ip = call_info.local_rtp_ip if call_info.local_rtp_ip else ""
            rtp_port = call_info.local_rtp_port

        _LOGGER.debug(
            "Starting RTP server on ip=%s, rtp_port=%s, rtcp_port=%s",
            rtp_ip,
            rtp_port,
            rtp_port + 1,
        )

        # Handle RTP packets in RTP server
        rtp_task = asyncio.create_task(
            self._create_rtp_server(protocol_factory, call_info, rtp_ip, rtp_port)
        )
        self._tasks.add(rtp_task)
        rtp_task.add_done_callback(self._tasks.remove)

        # Tell caller to start sending/receiving RTP audio
        self.answer(call_info, rtp_port)

    async def _create_rtp_server(
        self,
        protocol_factory: CallProtocolFactory,
        call_info: CallInfo,
        rtp_ip: str,
        rtp_port: int,
    ):
        # Shared state between RTP/RTCP servers
        rtcp_state = RtcpState()

        loop = asyncio.get_running_loop()

        # RTCP server
        await loop.create_datagram_endpoint(
            lambda: RtcpDatagramProtocol(rtcp_state),
            (rtp_ip, rtp_port + 1),
        )

        # RTP server
        await loop.create_datagram_endpoint(
            partial(protocol_factory, call_info, rtcp_state),
            (rtp_ip, rtp_port),
        )


class RtpDatagramProtocol(asyncio.DatagramProtocol, ABC):
    """Handle RTP audio input/output for a VoIP call."""

    def __init__(
        self,
        rate: int = 16000,
        width: int = 2,
        channels: int = 1,
        opus_payload_type: int = OPUS_PAYLOAD_TYPE,
        rtcp_state: Optional[RtcpState] = None,
    ) -> None:
        """Set up RTP server."""
        self.rtcp_state = rtcp_state

        if self.rtcp_state is not None:
            # Automatically disconnect when BYE is received over RTCP
            self.rtcp_state.bye_callback = self.disconnect

        # Desired format for input audio
        self.rate = rate
        self.width = width
        self.channels = channels

        self.transport = None
        self.addr = None

        self._audio_queue: "asyncio.Queue[bytes]" = asyncio.Queue()
        self._rtp_input = RtpOpusInput(opus_payload_type=opus_payload_type)
        self._rtp_output = RtpOpusOutput(opus_payload_type=opus_payload_type)
        self._is_connected: bool = False

    def disconnect(self):
        self._is_connected = False
        if self.transport is not None:
            self.transport.close()
            self.transport = None

    def connection_made(self, transport):
        """Server is ready."""
        self.transport = transport
        self._is_connected = True

    def datagram_received(self, data, addr):
        """Decode RTP + OPUS into raw audio."""
        if not self._is_connected:
            return

        self.addr = addr

        try:
            # STT expects 16Khz mono with 16-bit samples
            audio_bytes = self._rtp_input.process_packet(
                data,
                self.rate,
                self.width,
                self.channels,
            )

            self.on_chunk(audio_bytes)
        except Exception as err:
            self.disconnect()
            raise err

    @abstractmethod
    def on_chunk(self, audio_bytes: bytes) -> None:
        """Handle raw audio chunk."""

    def send_audio(
        self,
        audio_bytes: bytes,
        rate: int,
        width: int,
        channels: int,
        addr: Any = None,
        sleep_ratio: float = 1.0,
        silence_before: float = 0.0,
    ) -> None:
        """Send audio from WAV file in chunks over RTP."""
        if not self._is_connected:
            _LOGGER.debug("Not connected, can't send audio")
            return

        addr = addr or self.addr
        if addr is None:
            _LOGGER.debug("No destination address, can't send audio")
            raise ValueError("Destination address not set")

        bytes_per_sample = width * channels
        bytes_per_frame = self._rtp_output.opus_frame_size * bytes_per_sample

        # Generate all RTP packets up front
        sample_offset = 0
        samples_left = len(audio_bytes) // bytes_per_sample
        rtp_packets: list[bytes] = []
        while samples_left > 0:
            _LOGGER.debug("Preparing audio chunk to send")
            bytes_offset = sample_offset * bytes_per_sample
            chunk = audio_bytes[bytes_offset : bytes_offset + bytes_per_frame]
            samples_in_chunk = len(chunk) // bytes_per_sample
            samples_left -= samples_in_chunk

            for rtp_bytes in self._rtp_output.process_audio(
                chunk,
                rate,
                width,
                channels,
                is_end=samples_left <= 0,
            ):
                rtp_packets.append(rtp_bytes)

            sample_offset += samples_in_chunk

        # Pause before sending to allow time for user to pick up phone.
        _LOGGER.debug("Pause before sending")
        time.sleep(silence_before)

        # Send RTP in a steady stream, delaying between each packet to simulate real-time audio
        seconds_per_rtp = self._rtp_output.opus_frame_size / self._rtp_output.opus_rate
        for rtp_bytes in rtp_packets:
            if not self._is_connected:
                break

            if self.transport is not None:
                self.transport.sendto(rtp_bytes, addr)

            # Wait almost the full amount of time for the chunk.
            #
            # Sending too fast will cause the phone to skip chunks,
            # since it doesn't seem to have a very large buffer.
            #
            # Sending too slow will cause audio artifacts if there is
            # network jitter, which is why programs like GStreamer are
            # much better at this.
            time.sleep(seconds_per_rtp * sleep_ratio)


class RtcpDatagramProtocol(asyncio.DatagramProtocol, ABC):
    """UDP server for the Real-time Transport Control Protocol (RTCP)."""

    def __init__(self, state: RtcpState) -> None:
        """Set up RTCP server."""
        self.transport = None
        self.state = state
        self._is_connected = False

    def connection_made(self, transport):
        """Server ready."""
        self.transport = transport
        self._is_connected = True

    def disconnect(self):
        self._is_connected = False
        if self.transport is not None:
            self.transport.close()
            self.transport = None

    def datagram_received(self, data: bytes, addr):
        """Handle INVITE SIP messages."""
        if not self._is_connected:
            return

        try:
            if len(data) < 8:
                raise ValueError("RTCP packet is too small")

            # See: https://en.wikipedia.org/wiki/RTP_Control_Protocol#Packet_header
            _flags, packet_type, _packet_length, _ssrc = struct.unpack(
                ">BBHL", data[:8]
            )

            if packet_type == _RTCP_BYE:
                _LOGGER.debug("Received BYE message via RTCP from %s", addr)
                self.disconnect()

                if self.state.bye_callback is not None:
                    self.state.bye_callback()

        except Exception:
            _LOGGER.exception("Unexpected error handling RTCP packet")



"""Test voip_utils SIP functionality."""

from voip_utils.sip import SipEndpoint, get_sip_endpoint


def test_parse_header_for_uri():
    endpoint = SipEndpoint('"Test Name" <sip:12345@example.com>')
    assert endpoint.description == "Test Name"
    assert endpoint.uri == "sip:12345@example.com"
    assert endpoint.username == "12345"
    assert endpoint.host == "example.com"
    assert endpoint.port == 5060


def test_parse_header_for_uri_no_name():
    endpoint = SipEndpoint("sip:12345@example.com")
    assert endpoint.description is None
    assert endpoint.uri == "sip:12345@example.com"


def test_parse_header_for_uri_sips():
    endpoint = SipEndpoint('"Test Name" <sips:12345@example.com>')
    assert endpoint.description == "Test Name"
    assert endpoint.uri == "sips:12345@example.com"


def test_parse_header_for_uri_no_space_name():
    endpoint = SipEndpoint("Test <sip:12345@example.com>")
    assert endpoint.description == "Test"
    assert endpoint.uri == "sip:12345@example.com"


def test_parse_header_for_uri_no_space_between_name():
    endpoint = SipEndpoint("Test<sip:12345@example.com>")
    assert endpoint.description == "Test"
    assert endpoint.uri == "sip:12345@example.com"


def test_parse_header_for_uri_no_space_between_quoted_name():
    endpoint = SipEndpoint('"Test Endpoint"<sip:12345@example.com>')
    assert endpoint.description == "Test Endpoint"
    assert endpoint.uri == "sip:12345@example.com"


def test_parse_header_for_uri_no_username():
    endpoint = SipEndpoint("Test <sip:example.com>")
    assert endpoint.description == "Test"
    assert endpoint.username is None
    assert endpoint.uri == "sip:example.com"


def test_get_sip_endpoint():
    endpoint = get_sip_endpoint("example.com")
    assert endpoint.host == "example.com"
    assert endpoint.port == 5060
    assert endpoint.description is None
    assert endpoint.username is None
    assert endpoint.uri == "sip:example.com"


def test_get_sip_endpoint_with_username():
    endpoint = get_sip_endpoint("example.com", username="test")
    assert endpoint.host == "example.com"
    assert endpoint.port == 5060
    assert endpoint.description is None
    assert endpoint.username == "test"
    assert endpoint.uri == "sip:test@example.com"


def test_get_sip_endpoint_with_description():
    endpoint = get_sip_endpoint("example.com", description="Test Endpoint")
    assert endpoint.host == "example.com"
    assert endpoint.port == 5060
    assert endpoint.description == "Test Endpoint"
    assert endpoint.username is None
    assert endpoint.uri == "sip:example.com"
    assert endpoint.sip_header == '"Test Endpoint" <sip:example.com>'


def test_get_sip_endpoint_with_scheme():
    endpoint = get_sip_endpoint("example.com", scheme="sips")
    assert endpoint.host == "example.com"
    assert endpoint.port == 5060
    assert endpoint.description is None
    assert endpoint.username is None
    assert endpoint.uri == "sips:example.com"



"""Implementation of SIP (Session Initiation Protocol)."""

from __future__ import annotations

import asyncio
import logging
import re
import time
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Optional

from .const import OPUS_PAYLOAD_TYPE
from .error import VoipError
from .util import is_ipv4_address

SIP_PORT = 5060

_LOGGER = logging.getLogger(__name__)
_CRLF = "\r\n"

VOIP_UTILS_AGENT = "voip-utils"


@dataclass
class SdpInfo:
    """Information for Session Description Protocol (SDP)."""

    username: str
    id: int
    session_name: str
    version: str


@dataclass
class SipEndpoint:
    """Information about a SIP endpoint."""

    sip_header: str
    uri: str = field(init=False)
    scheme: str = field(init=False)
    host: str = field(init=False)
    port: int = field(init=False)
    username: str | None = field(init=False)
    description: str | None = field(init=False)

    def __post_init__(self):
        header_pattern = re.compile(
            r'\s*((?P<description>\b\w+\b|"[^"]+")\s*)?<?(?P<uri>sips?:[^>]+)>?.*'
        )
        header_match = header_pattern.match(self.sip_header)
        if header_match is not None:
            description_token = header_match.group("description")
            if description_token is not None:
                self.description = description_token.strip('"')
            else:
                self.description = None
            self.uri = header_match.group("uri")
            uri_pattern = re.compile(
                r"(?P<scheme>sips?):(?:(?P<user>[^@]+)@)?(?P<host>[^:;?]+)(?::(?P<port>\d+))?"
            )
            uri_match = uri_pattern.match(self.uri)
            if uri_match is None:
                raise ValueError("Invalid SIP uri")
            self.scheme = uri_match.group("scheme")
            self.username = uri_match.group("user")
            self.host = uri_match.group("host")
            self.port = (
                int(uri_match.group("port")) if uri_match.group("port") else SIP_PORT
            )
        else:
            raise ValueError("Invalid SIP header")


@dataclass
class SipMessage:
    """Data parsed from a SIP message."""

    protocol: str
    method: Optional[str]
    request_uri: Optional[str]
    code: Optional[str]
    reason: Optional[str]
    headers: dict[str, str]
    body: str

    @staticmethod
    def parse_sip(message: str) -> SipMessage:
        """Parse a SIP message into a SipMessage object."""
        lines = message.splitlines()

        method: Optional[str] = None
        request_uri: Optional[str] = None
        code: Optional[str] = None
        reason: Optional[str] = None
        headers: dict[str, str] = {}
        offset: int = 0

        # See: https://datatracker.ietf.org/doc/html/rfc3261
        for i, line in enumerate(lines):
            if line:
                offset += len(line) + len(_CRLF)

            if i == 0:
                line_parts = line.split()
                if line_parts[0].startswith("SIP"):
                    protocol = line_parts[0]
                    code = line_parts[1]
                    reason = line_parts[2]
                else:
                    method = line_parts[0]
                    request_uri = line_parts[1]
                    protocol = line_parts[2]
            elif not line:
                break
            else:
                key, value = line.split(":", maxsplit=1)
                headers[key.lower()] = value.strip()

        body = message[offset:]
        return SipMessage(protocol, method, request_uri, code, reason, headers, body)


@dataclass
class CallInfo:
    """Information gathered from an INVITE message."""

    caller_endpoint: SipEndpoint
    local_endpoint: SipEndpoint
    caller_rtp_port: int
    server_ip: str
    headers: dict[str, str]
    opus_payload_type: int = OPUS_PAYLOAD_TYPE
    local_rtp_ip: str | None = None
    local_rtp_port: int | None = None

    @property
    def caller_rtcp_port(self) -> int:
        """Real-time Transport Control Protocol (RTCP) port."""
        return self.caller_rtp_port + 1

    @property
    def caller_ip(self) -> str:
        """Get IP address of caller."""
        return self.caller_endpoint.host

    @property
    def caller_sip_port(self) -> int:
        """SIP port of caller."""
        return self.caller_endpoint.port

    @property
    def local_rtcp_port(self) -> int | None:
        """Get the local RTCP port."""
        return self.local_rtp_port + 1 if self.local_rtp_port is not None else None


@dataclass
class RtpInfo:
    """Information about the RTP transport used for the call audio."""

    rtp_ip: str | None
    rtp_port: int | None
    payload_type: int | None


def get_sip_endpoint(
    host: str,
    port: Optional[int] = None,
    scheme: Optional[str] = "sip",
    username: Optional[str] = None,
    description: Optional[str] = None,
) -> SipEndpoint:
    uri = f"{scheme}:"
    if username:
        uri += f"{username}@"
    uri += host
    if port:
        uri += f":{port}"
    if description:
        uri = f'"{description}" <{uri}>'
    return SipEndpoint(uri)


def get_rtp_info(body: str) -> RtpInfo:
    body_lines = body.splitlines()
    rtp_ip = None
    rtp_port = None
    opus_payload_type = None
    opus_payload_types_detected = []
    for line in body_lines:
        line = line.strip()
        if not line:
            continue

        key, _, value = line.partition("=")
        if key == "m":
            parts = value.split()
            if parts[0] == "audio":
                rtp_port = int(parts[1])
        elif key == "c":
            parts = value.split()
            if len(parts) > 2:
                rtp_ip = parts[2]
        elif key == "a" and value.startswith("rtpmap:"):
            # a=rtpmap:123 opus/48000/2
            codec_str = value.split(":", maxsplit=1)[1]
            codec_parts = codec_str.split()
            if (len(codec_parts) > 1) and (codec_parts[1].lower().startswith("opus")):
                opus_payload_types_detected.append(int(codec_parts[0]))
                _LOGGER.debug("Detected OPUS payload type as %s", opus_payload_type)

    if len(opus_payload_types_detected) > 0:
        opus_payload_type = opus_payload_types_detected[0]
        _LOGGER.debug("Using first detected payload type: %s", opus_payload_type)
    else:
        opus_payload_type = OPUS_PAYLOAD_TYPE
        _LOGGER.debug("Using default payload type: %s", opus_payload_type)

    return RtpInfo(rtp_ip=rtp_ip, rtp_port=rtp_port, payload_type=opus_payload_type)


class SipDatagramProtocol(asyncio.DatagramProtocol, ABC):
    """UDP server for the Session Initiation Protocol (SIP)."""

    def __init__(self, sdp_info: SdpInfo) -> None:
        """Set up SIP server."""
        self.sdp_info = sdp_info
        self.transport = None
        self._outgoing_calls: dict[str, int] = {}

    def outgoing_call(
        self, source: SipEndpoint, destination: SipEndpoint, rtp_port: int
    ) -> CallInfo:
        """Make an outgoing call from the given source endpoint to the destination endpoint, using the rtp_port for the local RTP port of the call."""
        if self.transport is None:
            raise RuntimeError("No transport available for outgoing call.")

        session_id = str(time.monotonic_ns())
        session_version = session_id
        call_id = session_id
        self._register_outgoing_call(call_id, rtp_port)

        sdp_lines = [
            "v=0",
            f"o={source.username} {session_id} {session_version} IN IP4 {source.host}",
            "s=Talk",
            f"c=IN IP4 {source.host}",
            "t=0 0",
            f"m=audio {rtp_port} RTP/AVP 123 96 101 103 104",
            "a=sendrecv",
            "a=rtpmap:96 opus/48000/2",
            "a=fmtp:96 useinbandfec=0",
            "a=rtpmap:123 opus/48000/2",
            "a=fmtp:123 maxplaybackrate=16000",
            "a=rtpmap:101 telephone-event/48000",
            "a=rtpmap:103 telephone-event/16000",
            "a=rtpmap:104 telephone-event/8000",
            "a=ptime:20",
            "",
        ]
        sdp_text = _CRLF.join(sdp_lines)
        sdp_bytes = sdp_text.encode("utf-8")

        invite_lines = [
            f"INVITE {destination.uri} SIP/2.0",
            f"Via: SIP/2.0/UDP {source.host}:{source.port}",
            f"From: {source.sip_header}",
            f"Contact: {source.sip_header}",
            f"To: {destination.sip_header}",
            f"Call-ID: {call_id}",
            "CSeq: 50 INVITE",
            f"User-Agent: {VOIP_UTILS_AGENT} 1.0",
            "Allow: INVITE, ACK, OPTIONS, CANCEL, BYE, SUBSCRIBE, NOTIFY, INFO, REFER, UPDATE",
            "Accept: application/sdp, application/dtmf-relay",
            "Content-Type: application/sdp",
            f"Content-Length: {len(sdp_bytes)}",
            "",
        ]
        invite_text = _CRLF.join(invite_lines) + _CRLF
        invite_bytes = invite_text.encode("utf-8")

        msg_bytes = invite_bytes + sdp_bytes

        _LOGGER.debug(msg_bytes)

        self.transport.sendto(
            msg_bytes,
            (destination.host, destination.port),
        )

        return CallInfo(
            caller_endpoint=destination,
            local_endpoint=source,
            caller_rtp_port=rtp_port,
            server_ip=source.host,
            headers={"call-id": call_id},
        )

    def hang_up(self, call_info: CallInfo):
        """Hang up the call when finished"""
        if self.transport is None:
            raise RuntimeError("No transport available for sending hangup.")

        bye_lines = [
            f"BYE {call_info.caller_endpoint.uri} SIP/2.0",
            f"Via: SIP/2.0/UDP {call_info.local_endpoint.host}:{call_info.local_endpoint.port}",
            f"From: {call_info.local_endpoint.sip_header}",
            f"To: {call_info.caller_endpoint.sip_header}",
            f"Call-ID: {call_info.headers['call-id']}",
            "CSeq: 51 BYE",
            f"User-Agent: {VOIP_UTILS_AGENT} 1.0",
            "Content-Length: 0",
            "",
        ]
        _LOGGER.debug("Hanging up...")
        bye_text = _CRLF.join(bye_lines) + _CRLF
        bye_bytes = bye_text.encode("utf-8")
        self.transport.sendto(
            bye_bytes, (call_info.caller_endpoint.host, call_info.caller_endpoint.port)
        )

        self._end_outgoing_call(call_info.headers["call-id"])
        self.on_hangup(call_info)

    def cancel_call(self, call_info: CallInfo):
        """Cancel an outgoing call while it's still ringing."""
        if self.transport is None:
            raise RuntimeError("No transport available for sending cancel.")

        cancel_lines = [
            f"CANCEL {call_info.caller_endpoint.uri} SIP/2.0",
            f"Via: SIP/2.0/UDP {call_info.local_endpoint.host}:{call_info.local_endpoint.port}",
            f"From: {call_info.local_endpoint.sip_header}",
            f"To: {call_info.caller_endpoint.sip_header}",
            f"Call-ID: {call_info.headers['call-id']}",
            "CSeq: 51 CANCEL",
            f"User-Agent: {VOIP_UTILS_AGENT} 1.0",
            "Content-Length: 0",
            "",
        ]
        _LOGGER.debug("Canceling call...")
        cancel_text = _CRLF.join(cancel_lines) + _CRLF
        cancel_bytes = cancel_text.encode("utf-8")
        self.transport.sendto(
            cancel_bytes,
            (call_info.caller_endpoint.host, call_info.caller_endpoint.port),
        )

        self._end_outgoing_call(call_info.headers["call-id"])
        self.on_hangup(call_info)

    def _register_outgoing_call(self, call_id: str, rtp_port: int):
        """Register the RTP port associated with an outgoing call."""
        self._outgoing_calls[call_id] = rtp_port

    def _get_call_rtp_port(self, call_id: str) -> int | None:
        """Get the RTP port associated with an outgoing call."""
        return self._outgoing_calls.get(call_id)

    def _end_outgoing_call(self, call_id: str):
        """Register the end of an outgoing call."""
        self._outgoing_calls.pop(call_id, None)

    def connection_made(self, transport):
        """Server ready."""
        self.transport = transport

    def datagram_received(self, data: bytes, addr):
        """Handle INVITE SIP messages."""
        try:
            if self.transport is None:
                _LOGGER.warning("No transport for exchanging SIP message")
                return

            caller_ip, caller_sip_port = addr
            message = data.decode("utf-8")
            smsg = SipMessage.parse_sip(message)
            _LOGGER.debug(
                "Received datagram protocol=[%s], method=[%s], ruri=[%s], code=[%s], reason=[%s], headers=[%s], body=[%s]",
                smsg.protocol,
                smsg.method,
                smsg.request_uri,
                smsg.code,
                smsg.reason,
                smsg.headers,
                smsg.body,
            )
            method = smsg.method
            if method is not None:
                method = method.lower()

            if method == "invite":
                # An invite message means someone called HA
                _LOGGER.debug("Received invite message")
                if not smsg.request_uri:
                    raise ValueError("Empty receiver URI")

                caller_endpoint = None
                # The From header should give us the URI used for sending SIP messages to the device
                if smsg.headers.get("from") is not None:
                    caller_endpoint = SipEndpoint(smsg.headers.get("from", ""))
                # We can try using the Contact header as a fallback
                elif smsg.headers.get("contact") is not None:
                    caller_endpoint = SipEndpoint(smsg.headers.get("contact", ""))
                # If all else fails try to generate a URI based on the IP and port from the address the message came from
                else:
                    caller_endpoint = get_sip_endpoint(caller_ip, port=caller_sip_port)

                local_endpoint = None
                if smsg.headers.get("to") is not None:
                    local_endpoint = SipEndpoint(smsg.headers.get("to", ""))
                else:
                    local_ip, local_port = self.transport.get_extra_info("sockname")
                    local_endpoint = get_sip_endpoint(local_ip, port=local_port)

                _LOGGER.debug("Incoming call from endpoint=%s", caller_endpoint)

                # Extract caller's RTP port from SDP.
                # See: https://datatracker.ietf.org/doc/html/rfc2327
                caller_rtp_port: Optional[int] = None
                opus_payload_type = OPUS_PAYLOAD_TYPE
                body_lines = smsg.body.splitlines()
                for line in body_lines:
                    line = line.strip()
                    if line:
                        key, value = line.split("=", maxsplit=1)
                        if key == "m":
                            parts = value.split()
                            if parts[0] == "audio":
                                caller_rtp_port = int(parts[1])
                        elif key == "a" and value.startswith("rtpmap:"):
                            # a=rtpmap:123 opus/48000/2
                            codec_str = value.split(":", maxsplit=1)[1]
                            codec_parts = codec_str.split()
                            if (len(codec_parts) > 1) and (
                                codec_parts[1].lower().startswith("opus")
                            ):
                                opus_payload_type = int(codec_parts[0])
                                _LOGGER.debug(
                                    "Detected OPUS payload type as %s",
                                    opus_payload_type,
                                )

                if caller_rtp_port is None:
                    raise VoipError("No caller RTP port")

                # Extract host from ruri
                # sip:user@123.123.123.123:1234
                re_splituri = re.compile(
                    r"(?P<scheme>\w+):"  # Scheme
                    + r"(?:(?P<user>[\w\.]+):?(?P<password>[\w\.]+)?@)?"  # User:Password
                    + r"\[?(?P<host>"  # Begin group host
                    + r"(?:\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})|"  # IPv4 address Host Or
                    + r"(?:(?:[0-9a-fA-F]{1,4}):){7}[0-9a-fA-F]{1,4}|"  # IPv6 address Host Or
                    + r"(?:(?:[0-9A-Za-z]+\.)+[0-9A-Za-z]+)"  # Hostname string
                    + r")\]?:?"  # End group host
                    + r"(?P<port>\d{1,6})?"  # port
                    + r"(?:\;(?P<params>[^\?]*))?"  # parameters
                    + r"(?:\?(?P<headers>.*))?"  # headers
                )
                re_uri = re_splituri.search(smsg.request_uri)
                if re_uri is None:
                    raise ValueError("Receiver URI did not match expected pattern")

                server_ip = re_uri.group("host")
                if not is_ipv4_address(server_ip):
                    raise VoipError(f"Invalid IPv4 address in {smsg.request_uri}")

                self.on_call(
                    CallInfo(
                        caller_endpoint=caller_endpoint,
                        local_endpoint=local_endpoint,
                        caller_rtp_port=caller_rtp_port,
                        server_ip=server_ip,
                        headers=smsg.headers,
                        opus_payload_type=opus_payload_type,
                    )
                )
            elif method is None:
                # Reply message means we must have received a response to someone we called
                # TODO: Verify that the call / sequence IDs match our outgoing INVITE
                _LOGGER.debug("Received response [%s]", message)
                is_ok = smsg.code == "200" and smsg.reason == "OK"
                if not is_ok:
                    _LOGGER.debug("Received non-OK response [%s]", message)
                    return

                _LOGGER.debug("Got OK message")
                if not self._is_response_type(smsg, "invite"):
                    # This will happen if/when we hang up.
                    _LOGGER.debug("Got response for non-invite message")
                    return

                _LOGGER.debug("Got invite response")
                rtp_info = get_rtp_info(smsg.body)
                remote_rtp_ip = rtp_info.rtp_ip
                remote_rtp_port = rtp_info.rtp_port
                opus_payload_type = rtp_info.payload_type
                caller_endpoint = None
                if smsg.headers.get("to") is not None:
                    caller_endpoint = SipEndpoint(smsg.headers.get("to", ""))
                else:
                    caller_endpoint = get_sip_endpoint(caller_ip, port=caller_sip_port)
                # The From header should give us the URI used for sending SIP messages to the device
                local_endpoint = None
                if smsg.headers.get("from") is not None:
                    local_endpoint = SipEndpoint(smsg.headers.get("from", ""))
                else:
                    local_endpoint = get_sip_endpoint(caller_ip, port=caller_sip_port)

                _LOGGER.debug("Outgoing call to endpoint=%s", caller_endpoint)
                ack_lines = [
                    f"ACK {caller_endpoint.uri} SIP/2.0",
                    f"Via: SIP/2.0/UDP {local_endpoint.host}:{local_endpoint.port}",
                    f"From: {local_endpoint.sip_header}",
                    f"To: {smsg.headers['to']}",
                    f"Call-ID: {smsg.headers['call-id']}",
                    "CSeq: 50 ACK",
                    f"User-Agent: {VOIP_UTILS_AGENT} 1.0",
                    "Content-Length: 0",
                ]
                ack_text = _CRLF.join(ack_lines) + _CRLF
                ack_bytes = ack_text.encode("utf-8")
                self.transport.sendto(ack_bytes, (caller_ip, caller_sip_port))

                # The call been answered, proceed with desired action here
                local_rtp_port = self._get_call_rtp_port(smsg.headers["call-id"])
                self.on_call(
                    CallInfo(
                        caller_endpoint=caller_endpoint,
                        local_endpoint=local_endpoint,
                        caller_rtp_port=remote_rtp_port,
                        server_ip=remote_rtp_ip,
                        headers=smsg.headers,
                        opus_payload_type=opus_payload_type,  # Should probably update this to eventually support more codecs
                        local_rtp_ip=local_endpoint.host,
                        local_rtp_port=local_rtp_port,
                    )
                )
            elif method == "bye":
                # Acknowlege the BYE message when the remote party hangs up
                _LOGGER.debug("Received BYE message: %s", message)
                if self.transport is None:
                    _LOGGER.debug("Skipping message: %s", message)
                    return

                # The From header should give us the URI used for sending SIP messages to the device
                if smsg.headers.get("from") is not None:
                    caller_endpoint = SipEndpoint(smsg.headers.get("from", ""))
                # We can try using the Contact header as a fallback
                elif smsg.headers.get("contact") is not None:
                    caller_endpoint = SipEndpoint(smsg.headers.get("contact", ""))
                # If all else fails try to generate a URI based on the IP and port from the address the message came from
                else:
                    caller_endpoint = get_sip_endpoint(caller_ip, port=caller_sip_port)

                local_endpoint = None
                if smsg.headers.get("to") is not None:
                    local_endpoint = SipEndpoint(smsg.headers.get("to", ""))
                else:
                    local_ip, local_port = self.transport.get_extra_info("sockname")
                    local_endpoint = get_sip_endpoint(local_ip, port=local_port)

                _LOGGER.debug("Incoming BYE from endpoint=%s", caller_endpoint)

                # Acknowledge the BYE message, otherwise the phone will keep sending it
                rtp_info = get_rtp_info(smsg.body)
                remote_rtp_ip = rtp_info.rtp_ip
                remote_rtp_port = rtp_info.rtp_port
                opus_payload_type = rtp_info.payload_type
                # We should remove the call from the outgoing calls dict now if it is there
                self._end_outgoing_call(smsg.headers["call-id"])
                ok_lines = [
                    "SIP/2.0 200 OK",
                    f"Via: {smsg.headers['via']}",
                    f"From: {smsg.headers['from']}",
                    f"To: {smsg.headers['to']}",
                    f"Call-ID: {smsg.headers['call-id']}",
                    f"CSeq: {smsg.headers['cseq']}",
                    f"User-Agent: {VOIP_UTILS_AGENT} 1.0",
                    "Content-Length: 0",
                ]
                ok_text = _CRLF.join(ok_lines) + _CRLF
                ok_bytes = ok_text.encode("utf-8")
                # We should probably tell the associated RTP server to shutdown at this point, assuming we aren't reusing it for other calls
                _LOGGER.debug("Sending OK for BYE message: %s", ok_text)
                self.transport.sendto(
                    ok_bytes,
                    (caller_ip, caller_sip_port),
                )
                # The transport might be used for incoming calls
                # as well, so we should leave it open.

                # Cleanup any necessary call state
                self.on_hangup(
                    CallInfo(
                        caller_endpoint=caller_endpoint,
                        local_endpoint=local_endpoint,
                        caller_rtp_port=remote_rtp_port,
                        server_ip=remote_rtp_ip,
                        headers=smsg.headers,
                    )
                )

        except Exception:
            _LOGGER.exception("Unexpected error handling SIP message")

    @abstractmethod
    def on_call(self, call_info: CallInfo):
        """Handle incoming calls."""

    def on_hangup(self, call_info: CallInfo):
        """Handle the end of a call."""

    def _is_response_type(self, msg: SipMessage, resp_type: str) -> bool:
        """Return whether or not the response message is for the given type."""
        return (
            msg is not None
            and "cseq" in msg.headers
            and resp_type.lower() in msg.headers["cseq"].lower()
        )

    def answer(
        self,
        call_info: CallInfo,
        server_rtp_port: int,
    ):
        """Send OK message to caller with our IP and RTP port."""
        if self.transport is None:
            return

        # SDP = Session Description Protocol
        # See: https://datatracker.ietf.org/doc/html/rfc2327
        body_lines = [
            "v=0",
            f"o={self.sdp_info.username} {self.sdp_info.id} 1 IN IP4 {call_info.server_ip}",
            f"s={self.sdp_info.session_name}",
            f"c=IN IP4 {call_info.server_ip}",
            "t=0 0",
            f"m=audio {server_rtp_port} RTP/AVP {call_info.opus_payload_type}",
            f"a=rtpmap:{call_info.opus_payload_type} opus/48000/2",
            "a=ptime:20",
            "a=maxptime:150",
            "a=sendrecv",
            _CRLF,
        ]
        body = _CRLF.join(body_lines)

        response_headers = {
            "Via": call_info.headers["via"],
            "From": call_info.headers["from"],
            "To": call_info.headers["to"],
            "Call-ID": call_info.headers["call-id"],
            "Content-Type": "application/sdp",
            "Content-Length": len(body),
            "CSeq": call_info.headers["cseq"],
            "Contact": call_info.headers["contact"],
            "User-Agent": f"{self.sdp_info.username} {self.sdp_info.id} {self.sdp_info.version}",
            "Allow": "INVITE, ACK, BYE, CANCEL, OPTIONS",
        }
        response_lines = ["SIP/2.0 200 OK"]

        for key, value in response_headers.items():
            response_lines.append(f"{key}: {value}")

        response_lines.append(_CRLF)
        response_str = _CRLF.join(response_lines) + body
        response_bytes = response_str.encode()

        self.transport.sendto(
            response_bytes,
            (call_info.caller_ip, call_info.caller_sip_port),
        )
        _LOGGER.debug(
            "Sent OK to ip=%s, port=%s with rtp_port=%s",
            call_info.caller_ip,
            call_info.caller_sip_port,
            server_rtp_port,
        )



"""Utilities for VoIP"""
from ipaddress import IPv4Address


def is_ipv4_address(address: str) -> bool:
    """Check if a given string is an IPv4 address."""
    try:
        IPv4Address(address)
    except ValueError:
        return False

    return True



"""Voice over IP utilities."""

from .sip import SIP_PORT, CallInfo, SdpInfo, SipDatagramProtocol
from .voip import (
    RtcpDatagramProtocol,
    RtcpState,
    RtpDatagramProtocol,
    VoipDatagramProtocol,
)



import asyncio
import logging
import os
import socket
from functools import partial
from pathlib import Path
from typing import Any, Callable, Optional, Set

from dotenv import load_dotenv

from voip_utils.voip import VoipDatagramProtocol, CallProtocolFactory
from voip_utils.sip import (
    CallInfo,
    SdpInfo,
    SipEndpoint,
    get_sip_endpoint,
)
from voip_utils.voip import RtcpDatagramProtocol, RtcpState, RtpDatagramProtocol

_LOGGER = logging.getLogger(__name__)

load_dotenv()


def get_env_int(env_var: str, default_val: int) -> int:
    value = os.getenv(env_var)
    if value is None:
        return default_val
    try:
        return int(value)
    except ValueError:
        return default_val


CALL_SRC_USER = os.getenv("CALL_SRC_USER")
CALL_SRC_IP = os.getenv("CALL_SRC_IP", "127.0.0.1")
CALL_SRC_PORT = get_env_int("CALL_SRC_PORT", 5060)
CALL_VIA_IP = os.getenv("CALL_VIA_IP")
CALL_DEST_IP = os.getenv("CALL_DEST_IP", "127.0.0.1")
CALL_DEST_PORT = get_env_int("CALL_DEST_PORT", 5060)
CALL_DEST_USER = os.getenv("CALL_DEST_USER")


RATE = 16000
WIDTH = 2
CHANNELS = 1
RTP_AUDIO_SETTINGS = {
    "rate": RATE,
    "width": WIDTH,
    "channels": CHANNELS,
    "sleep_ratio": 0.99,
}

class ExampleVoipDatagramProtocol(VoipDatagramProtocol):

    def __init__(
        self,
        sdp_info: SdpInfo,
        valid_protocol_factory: CallProtocolFactory,
        invalid_protocol_factory: Optional[CallProtocolFactory] = None,
    ) -> None:
        """Set up VoIP call handler."""
        super().__init__(sdp_info, valid_protocol_factory, invalid_protocol_factory)
        self.call_end = asyncio.Event()


    def on_hangup(self, call_info: CallInfo):
        """Example implementation of on hangup. """
        self.call_end.set()

class PreRecordMessageProtocol(RtpDatagramProtocol):
    """Plays a pre-recorded message on a loop."""

    def __init__(
        self,
        file_name: str,
        opus_payload_type: int,
        message_delay: float = 1.0,
        loop_delay: float = 2.0,
        rtcp_state: RtcpState | None = None,
    ) -> None:
        """Set up RTP server."""
        super().__init__(
            rate=RATE,
            width=WIDTH,
            channels=CHANNELS,
            opus_payload_type=opus_payload_type,
            rtcp_state=rtcp_state,
        )
        self.loop = asyncio.get_running_loop()
        self.file_name = file_name
        self.message_delay = message_delay
        self.loop_delay = loop_delay
        self._audio_task: asyncio.Task | None = None
        file_path = Path(__file__).parent / self.file_name
        self._audio_bytes: bytes = file_path.read_bytes()
        _LOGGER.debug("Created PreRecordMessageProtocol")

    def on_chunk(self, audio_bytes: bytes) -> None:
        """Handle raw audio chunk."""
        _LOGGER.debug("on_chunk")
        if self.transport is None:
            return

        if self._audio_task is None:
            self._audio_task = self.loop.create_task(
                self._play_message(),
                name="voip_not_connected",
            )

    async def _play_message(self) -> None:
        _LOGGER.debug("_play_message")
        self.send_audio(
            self._audio_bytes,
            self.rate,
            self.width,
            self.channels,
            self.addr,
            silence_before=self.message_delay,
        )

        await asyncio.sleep(self.loop_delay)

        # Allow message to play again - Only play once for testing
        # self._audio_task = None


async def main() -> None:
    logging.basicConfig(level=logging.DEBUG)

    loop = asyncio.get_event_loop()
    source = get_sip_endpoint(
        host=CALL_SRC_IP, port=CALL_SRC_PORT, username=CALL_SRC_USER, description=None
    )
    destination = get_sip_endpoint(
        host=CALL_DEST_IP,
        port=CALL_DEST_PORT,
        username=CALL_DEST_USER,
        description=None,
    )

    # Find free RTP/RTCP ports
    rtp_port = 0

    while True:
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.setblocking(False)

        # Bind to a random UDP port
        sock.bind(("", 0))
        _, rtp_port = sock.getsockname()

        # Close socket to free port for re-use
        sock.close()

        # Check that the next port up is available for RTCP
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        try:
            sock.bind(("", rtp_port + 1))

            # Will be opened again below
            sock.close()

            # Found our ports
            break
        except OSError:
            # RTCP port is taken
            pass

    _, protocol = await loop.create_datagram_endpoint(
        lambda: ExampleVoipDatagramProtocol(
            None,
            lambda call_info, rtcp_state: PreRecordMessageProtocol(
                "problem.pcm",
                call_info.opus_payload_type,
                rtcp_state=rtcp_state,
            ),
            lambda call_info, rtcp_state: PreRecordMessageProtocol(
                "problem.pcm",
                call_info.opus_payload_type,
                rtcp_state=rtcp_state,
            ),
        ),
        local_addr=(CALL_SRC_IP, CALL_SRC_PORT),
    )

    protocol.outgoing_call(source, destination, rtp_port)

    await protocol.call_end.wait()


if __name__ == "__main__":
    asyncio.run(main())



"""Utility for converting audio to/from RTP + OPUS packets."""

try:
    # Use built-in audioop until it's removed in Python 3.13
    import audioop  # pylint: disable=deprecated-module
except ImportError:
    from . import pyaudioop as audioop  # type: ignore[no-redef]

import logging
import random
import struct
from collections.abc import Iterable
from dataclasses import dataclass
from typing import Any

import opuslib

from .const import OPUS_PAYLOAD_TYPE
from .error import RtpError

_LOGGER = logging.getLogger(__name__)


@dataclass
class RtpOpusInput:
    """Extracts audio from RTP packets with OPUS."""

    opus_rate: int = 48000  # Hz
    opus_width: int = 2  # bytes
    opus_channels: int = 2
    opus_frame_size: int = 960  # samples per channel
    opus_payload_type: int = OPUS_PAYLOAD_TYPE  # set by GrandStream

    def __post_init__(
        self,
    ) -> None:
        """Initialize encoder and state."""
        self._decoder = opuslib.api.decoder.create_state(
            self.opus_rate, self.opus_channels
        )

    def process_packet(
        self,
        rtp_bytes: bytes,
        rate: int,
        width: int,
        channels: int,
    ) -> bytes:
        """Extract, decode, and return raw audio from RTP packet."""
        if channels not in (1, 2):
            raise ValueError("Only mono and stereo audio is supported")

        # Minimum header size
        if len(rtp_bytes) < 12:
            raise RtpError("RTP packet is too small")

        # See: https://en.wikipedia.org/wiki/Real-time_Transport_Protocol#Packet_header
        flags, payload_type, _sequence_num, _timestamp, _ssrc = struct.unpack(
            ">BBHLL", rtp_bytes[:12]
        )

        if flags != 0b10000000:
            raise RtpError("Padding and extension headers not supported")

        payload_type &= 0x7F  # Remove marker bit
        if payload_type != self.opus_payload_type:
            raise RtpError(
                f"Expected payload type {self.opus_payload_type}, got {payload_type}"
            )

        # Assume no padding, extension headers, etc.
        opus_bytes = rtp_bytes[12:]

        # Decode into raw audio.
        # This will always be 48Khz stereo with 16-bit samples.
        audio_bytes = opuslib.api.decoder.decode(
            self._decoder,
            opus_bytes,
            len(opus_bytes),
            self.opus_frame_size,
            False,  # no forward error correction (fec)
        )

        # Convert to target sample rate, etc.
        if channels == 1:
            # Convert to mono
            audio_bytes = audioop.tomono(
                audio_bytes,
                self.opus_width,
                1.0,
                1.0,
            )

        if rate != self.opus_rate:
            # Resample
            audio_bytes, _state = audioop.ratecv(
                audio_bytes,
                self.opus_width,
                channels,
                self.opus_rate,
                rate,
                None,
            )

        if width != self.opus_width:
            # Resize
            audio_bytes = audioop.lin2lin(
                audio_bytes,
                self.opus_width,
                width,
            )

        return audio_bytes


@dataclass
class RtpOpusOutput:
    """Prepares audio to send to an RTP client using OPUS."""

    opus_rate: int = 48000  # Hz
    opus_width: int = 2  # bytes
    opus_channels: int = 2
    opus_frame_size: int = 960  # samples per channel
    opus_payload_type: int = OPUS_PAYLOAD_TYPE  # set by GrandStream
    opus_bytes_per_frame: int = 960 * 2 * 2  # 16-bit x stereo

    _rtp_flags: int = 0b10000000  # v2, no padding/extensions/CSRCs
    _rtp_sequence_num: int = 0
    _rtp_timestamp: int = 0
    _rtp_ssrc: int = 0

    _encoder: opuslib.api.encoder.Encoder = None
    _audio_buffer: bytes = None  # type: ignore[assignment]
    _resample_state: Any = None

    def __post_init__(
        self,
    ) -> None:
        """Initialize encoder and state."""
        self.opus_bytes_per_frame = (
            self.opus_frame_size * self.opus_width * self.opus_channels
        )

        # Set up OPUS encoder for VoIP
        self._encoder = opuslib.api.encoder.create_state(
            self.opus_rate,
            self.opus_width,
            opuslib.APPLICATION_VOIP,
        )

        self.reset()

    def reset(self):
        """Clear audio buffer and state."""
        self._audio_buffer = b""
        self._resample_state = None

        # Recommended to start from random offsets to aid encryption
        self._rtp_sequence_num = random.randint(0, 2**10)
        self._rtp_timestamp = random.randint(1, 2**10)

        # Change each time
        self._rtp_ssrc = random.randint(0, 2**32)

    def process_audio(
        self,
        audio_bytes: bytes,
        rate: int,
        width: int,
        channels: int,
        is_end: bool = False,
    ) -> Iterable[bytes]:
        """Process a chunk of raw audio and yield RTP packet(s)."""
        if rate != self.opus_rate:
            # Convert to 48Khz
            audio_bytes, self._resample_state = audioop.ratecv(
                audio_bytes,
                width,
                channels,
                rate,
                self.opus_rate,
                self._resample_state,
            )

        if width != self.opus_width:
            # Adjust sample width
            audio_bytes = audioop.lin2lin(
                audio_bytes,
                width,
                self.opus_width,
            )

        if channels != self.opus_channels:
            # Convert to stereo
            audio_bytes = audioop.tostereo(
                audio_bytes,
                self.opus_width,
                1.0,
                1.0,
            )

        self._audio_buffer += audio_bytes
        if is_end:
            # Pad with silence
            bytes_missing = len(self._audio_buffer) % self.opus_bytes_per_frame
            if bytes_missing > 0:
                self._audio_buffer += bytes(bytes_missing)

        num_frames = len(self._audio_buffer) // self.opus_bytes_per_frame

        # Process chunks with *exactly* the desired number of frames
        for i in range(num_frames):
            offset = i * self.opus_bytes_per_frame
            audio_chunk = self._audio_buffer[
                offset : offset + self.opus_bytes_per_frame
            ]

            # Encode to OPUS packet
            opus_bytes = opuslib.api.encoder.encode(
                self._encoder,
                audio_chunk,
                self.opus_frame_size,
                4000,  # recommended in opus docs
            )

            # Add RTP header
            # See: https://en.wikipedia.org/wiki/Real-time_Transport_Protocol#Packet_header
            rtp_bytes = struct.pack(
                ">BBHLL",
                self._rtp_flags,
                self.opus_payload_type,
                self._rtp_sequence_num,
                self._rtp_timestamp,
                self._rtp_ssrc,
            )

            # RTP packet
            yield rtp_bytes + opus_bytes

            # Next frame
            self._rtp_sequence_num += 1
            self._rtp_timestamp += self.opus_frame_size

        if num_frames > 0:
            # Remove audio already sent
            self._audio_buffer = self._audio_buffer[
                num_frames * self.opus_bytes_per_frame :
            ]

        if is_end:
            # Clear audio buffer and state
            self.reset()
